@@ -18,7 +18,7 @@ if __name__ == '__main__':
     #cProfile.run('pytest.main([%r] + sys.argv[1:])' % __file__,
     #             'cProfile.dat')
     #sys.exit()
-    
+
     sys.exit(pytest.main([__file__] + sys.argv[1:]))
 
 from dugong import HTTPConnection, BodyFollowing, CaseInsensitiveDict, _join
@@ -36,6 +36,7 @@ import os
 import hashlib
 import threading
 import socketserver
+from pytest import raises as assert_raises
 
 # We want to test with a real certificate
 SSL_TEST_HOST = 'www.google.com'
@@ -116,7 +117,7 @@ def check_http_connection():
 
 def test_connect_ssl():
     check_http_connection()
-    
+
     ssl_context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
     ssl_context.options |= ssl.OP_NO_SSLv2
     ssl_context.verify_mode = ssl.CERT_REQUIRED
@@ -137,7 +138,7 @@ def test_invalid_ssl():
     context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
     context.options |= ssl.OP_NO_SSLv2
     context.verify_mode = ssl.CERT_REQUIRED
-    
+
     conn = HTTPConnection(SSL_TEST_HOST, ssl_context=context)
     with pytest.raises(ssl.SSLError):
         conn.send_request('GET', '/')
@@ -149,7 +150,7 @@ def test_get_pipeline(conn):
     # a few requests
 
     paths = [ '/send_120_bytes' for _ in range(3) ]
-    
+
     # Send requests
     for path in paths:
         crt = conn.co_send_request('GET', path)
@@ -167,7 +168,7 @@ def test_get_pipeline(conn):
 def test_ssl_info(conn):
     conn.get_ssl_cipher()
     conn.get_ssl_peercert()
-    
+
 def test_blocking_send(conn):
     # Send requests until we block because all TCP buffers are full
 
@@ -198,7 +199,7 @@ def test_blocking_send(conn):
     resp = conn.read_response()
     assert resp.status == 200
     conn.discard()
-        
+
 def test_blocking_read(conn):
     delay = 10
     while True:
@@ -221,7 +222,7 @@ def test_blocking_read(conn):
                     break
                 parts.append(buf)
         assert not conn.response_pending()
-        
+
         assert _join(parts) == DUMMY_DATA[:120]*10
         if interrupted >= 8:
             break
@@ -246,7 +247,7 @@ def test_discard_chunked(conn):
     assert resp.length is None
     conn.discard()
     assert not conn.response_pending()
-    
+
 def test_read_text(conn):
     conn.send_request('GET', '/send_%d_bytes' % len(DUMMY_DATA))
     conn.read_response()
@@ -258,7 +259,7 @@ def test_read_text2(conn):
     conn.send_request('GET', '/send_%d_bytes' % len(DUMMY_DATA))
     conn.read_response()
     fh = TextIOWrapper(conn)
-    
+
     # This used to fail because TextIOWrapper can't deal with bytearrays
     fh.read(42)
 
@@ -266,7 +267,7 @@ def test_read_text3(conn):
     conn.send_request('GET', '/send_%d_bytes' % len(DUMMY_DATA))
     conn.read_response()
     fh = TextIOWrapper(conn)
-    
+
     # This used to fail because TextIOWrapper tries to read from
     # the underlying fh even after getting ''
     while True:
@@ -274,7 +275,7 @@ def test_read_text3(conn):
             break
 
     assert not conn.response_pending()
-    
+
 def test_read_identity(conn):
     conn.send_request('GET', '/send_512_bytes')
     resp = conn.read_response()
@@ -287,11 +288,11 @@ def test_read_identity(conn):
 def test_exhaust_buffer(conn):
     if conn.ssl_context:
         pytest.skip('test does not have ssl support yet')
-        
+
     conn._rbuf = dugong._Buffer(600)
     conn.send_request('GET', '/send_512_bytes')
     conn.read_response()
-    
+
     # Test the case where the readbuffer is truncated and
     # returned, instead of copied
     conn._rbuf.compact()
@@ -307,7 +308,7 @@ def test_exhaust_buffer(conn):
 def test_full_buffer(conn):
     if conn.ssl_context:
         pytest.skip('test does not have ssl support yet')
-        
+
     conn._rbuf = dugong._Buffer(100)
     conn.send_request('GET', '/send_512_bytes')
     conn.read_response()
@@ -320,9 +321,9 @@ def test_full_buffer(conn):
     assert conn._rbuf.e == 0
     conn._rbuf.e = len(conn._rbuf.d)
     conn._rbuf.b = conn._rbuf.e
-    
+
     assert conn.readall() == DUMMY_DATA[pos:512]
-    
+
 def test_readinto_identity(conn):
     conn.send_request('GET', '/send_512_bytes')
     resp = conn.read_response()
@@ -338,7 +339,7 @@ def test_readinto_identity(conn):
         parts.append(buf[:len_])
     assert _join(parts) == DUMMY_DATA[:512]
     assert not conn.response_pending()
-    
+
 def test_read_chunked(conn):
     conn.send_request('GET', '/send_3_300-byte_chunks')
     resp = conn.read_response()
@@ -347,6 +348,14 @@ def test_read_chunked(conn):
     assert conn.readall() == DUMMY_DATA[:300]*3
     assert not conn.response_pending()
 
+def test_read_chunked2(conn):
+    conn.send_request('GET', '/send_10_5-byte_chunks')
+    resp = conn.read_response()
+    assert resp.status == 200
+    assert resp.length is None
+    assert conn.readall() == DUMMY_DATA[:5]*10
+    assert not conn.response_pending()
+
 def test_readinto_chunked(conn):
     conn.send_request('GET', '/send_3_300-byte_chunks')
     resp = conn.read_response()
@@ -362,7 +371,7 @@ def test_readinto_chunked(conn):
         parts.append(buf[:len_])
     assert _join(parts) == DUMMY_DATA[:300] * 3
     assert not conn.response_pending()
-    
+
 def test_double_read(conn):
     conn.send_request('GET', '/send_10_bytes')
     resp = conn.read_response()
@@ -380,7 +389,30 @@ def test_read_raw(conn):
         conn.readall()
     assert conn.read_raw(512) == b'body data'
     assert conn.read_raw(512) == b''
-        
+
+def test_abort_read(conn):
+    conn.send_request('GET', '/send_3_300-byte_chunks')
+    resp = conn.read_response()
+    assert resp.status == 200
+    conn.read(200)
+    conn.disconnect()
+    assert_raises(dugong.ConnectionClosed, conn.read, 200)
+
+def test_abort_co_read(conn):
+    conn.send_request('GET', '/send_3_300-byte_chunks')
+    resp = conn.read_response()
+    assert resp.status == 200
+    cofun = conn.co_read(450)
+    next(cofun)
+    conn.disconnect()
+    assert_raises(dugong.ConnectionClosed, next, cofun)
+
+def test_abort_write(conn):
+    conn.send_request('PUT', '/allgood', body=BodyFollowing(42))
+    conn.write(b'fooo')
+    conn.disconnect()
+    assert_raises(dugong.ConnectionClosed, conn.write, b'baar')
+
 def test_write_toomuch(conn):
     conn.send_request('PUT', '/allgood', body=BodyFollowing(42))
     with pytest.raises(dugong.ExcessBodyData):
@@ -481,7 +513,7 @@ def test_100cont_2(conn):
 
     conn.read_response()
     conn.readall()
-    
+
 def test_100cont_3(conn):
     conn.send_request('PUT', '/fail_with_403', body=BodyFollowing(256), expect100=True)
 
@@ -490,11 +522,11 @@ def test_100cont_3(conn):
 
     conn.read_response()
     conn.readall()
-    
+
 def test_tunnel(http_server):
     if http_server.use_ssl:
         pytest.skip('test does not have ssl support yet')
-        
+
     conn = HTTPConnection('remote_server', proxy=(http_server.host, http_server.port))
 
     conn.send_request('GET', '/send_10_bytes')
@@ -565,7 +597,7 @@ def test_smallbuffer(conn, buffer_size):
     assert resp.length == 512
     assert conn.readall() == DUMMY_DATA[:512]
     assert not conn.response_pending()
-    
+
 def test_mutable_read(conn):
     # Read data and modify it, to make sure that this doesn't
     # affect the buffer
@@ -593,7 +625,7 @@ def test_mutable_read(conn):
 
     assert conn.readall() == DUMMY_DATA[pos:512]
     assert not conn.response_pending()
-    
+
 
 DUMMY_DATA = ','.join(str(x) for x in range(10000)).encode()
 
@@ -604,7 +636,7 @@ class MockRequestHandler(BaseHTTPRequestHandler):
 
     #def log_message(self, format, *args):
     #    pass
-    
+
     def handle_expect_100(self):
         if self.handle_errors():
             return
@@ -623,6 +655,8 @@ class MockRequestHandler(BaseHTTPRequestHandler):
                 pass
             else:
                 raise
+        except BrokenPipeError:
+            pass
 
     def do_GET(self):
         if self.handle_errors():
@@ -640,7 +674,7 @@ class MockRequestHandler(BaseHTTPRequestHandler):
             self.wfile.write(b'body data')
             self.wfile.close()
             return
-            
+
         hit = re.match(r'^/send_([0-9]+)_bytes', self.path)
         if hit:
             len_ = int(hit.group(1))
@@ -676,6 +710,7 @@ class MockRequestHandler(BaseHTTPRequestHandler):
                 if i % 3 == 2 and delay:
                     time.sleep(delay)
                 self.wfile.write(b'\r\n')
+                self.wfile.flush()
             self.wfile.write(b'0\r\n\r\n')
             return
 
